from modelgauge.prompt import ChatPrompt, ChatRole


def format_chat(
    chat: ChatPrompt, *, user_role: str = "user", sut_role: str = "assistant"
) -> str:
    """Flattens a chat conversation into a single text prompt"""
    blocks = []
    for message in chat.messages:
        role_text: str
        if message.role == ChatRole.user:
            role_text = user_role
        else:
            role_text = sut_role
        blocks.append(f"{role_text}: {message.text}")
    blocks.append(f"{sut_role}: ")
    return "\n\n".join(blocks)



from functools import wraps


def not_implemented(f):
    """Decorate a method as not implemented in a way we can detect."""

    @wraps(f)
    def inner(*args, **kwargs):
        f(*args, **kwargs)
        # We expect the previous line to raise a NotImplementedError, assert if it doesn't
        raise AssertionError(f"Expected {f} to raise a NotImplementedError.")

    inner._not_implemented = True
    return inner


def is_not_implemented(f) -> bool:
    """Check if a method is decorated with @not_implemented."""
    return getattr(f, "_not_implemented", False)



"""
This module supports the common pattern of structuring Prompts
as multiple choice questions.
"""
import random
from collections import defaultdict
from dataclasses import dataclass
from itertools import cycle
from typing import Dict, List, Optional


@dataclass(frozen=True)
class MultipleChoiceQuestion:
    """Hold all the data relevant to a single question."""

    question: str
    options: List[str]
    correct_option: int  # Index into options


@dataclass(frozen=True, kw_only=True)
class MultipleChoiceFormatting:
    """Options for how to convert MultipleChoiceQuestion into Prompt text."""

    question_prefix: str = ""
    question_suffix: str = ""
    option_identifiers: List[str]
    option_identifier_separator: str
    option_separator: str
    answer_prefix: str
    answer_suffix: str = ""
    # Used when combining multiple questions together.
    instructions_block: str = ""
    block_separator: str = ""


def basic_multiple_choice_format() -> MultipleChoiceFormatting:
    """A standard setting for how to format multiple choice questions."""
    return MultipleChoiceFormatting(
        question_prefix="Question: ",
        question_suffix="\n",
        # All capital letters
        option_identifiers=[chr(ord("A") + i) for i in range(26)],
        option_identifier_separator=") ",
        option_separator="\n",
        answer_prefix="\nAnswer: ",
        answer_suffix="\n",
        instructions_block="The following are multiple choice questions (with answers).",
        block_separator="\n",
    )


def question_with_training_to_text(
    eval_question: MultipleChoiceQuestion,
    training_questions: List[MultipleChoiceQuestion],
    formatting: MultipleChoiceFormatting,
) -> str:
    """Creates a series of multiple choice question blocks."""

    blocks: List[str] = []
    if formatting.instructions_block:
        blocks.append(formatting.instructions_block)

    # Text for in-context training questions
    for train in training_questions:
        blocks.append(question_to_text(train, formatting, include_answer=True))

    # The eval question's text
    blocks.append(question_to_text(eval_question, formatting, include_answer=False))

    return formatting.block_separator.join(blocks)


def question_to_text(
    question: MultipleChoiceQuestion,
    formatting: MultipleChoiceFormatting,
    include_answer: bool,
) -> str:
    """Formats a multiple choice question."""
    # Start with the question
    result: str = (
        formatting.question_prefix + question.question + formatting.question_suffix
    )

    # Add each option
    option_blocks = []
    for option_index, option in enumerate(question.options):
        identifier = formatting.option_identifiers[option_index]
        option_blocks.append(
            identifier + formatting.option_identifier_separator + option
        )

    result += formatting.option_separator.join(option_blocks)

    # Either include the answer or the prefix to the answer.
    if include_answer:
        correct_identifier = formatting.option_identifiers[question.correct_option]
        result += (
            formatting.answer_prefix + correct_identifier + formatting.answer_suffix
        )
    else:
        result += formatting.answer_prefix.rstrip()

    return result


def sample_examples(
    all_train_questions: List[MultipleChoiceQuestion], seed: int, train_sample_size: int
) -> List[MultipleChoiceQuestion]:
    """This is NOT supported, it is just to for backward comparability with HELM.

    This is a copy of HELM's InContextLearningAdapter.sample_examples minimally updated.
    It includes the odd behavior reported in HELM's Issue #2224."""
    # Fix the random seed for reproducibility
    rng = random.Random()
    rng.seed(seed)
    num_instances_to_sample: int = min(len(all_train_questions), train_sample_size)

    examples: List[MultipleChoiceQuestion] = []
    label_to_instances: Dict[str, List[MultipleChoiceQuestion]] = defaultdict(list)
    for question in all_train_questions:
        label_to_instances[question.options[question.correct_option]].append(question)

    # Build counts to labels
    instances: List[MultipleChoiceQuestion]
    counts_to_labels: Dict[int, List[str]] = defaultdict(list)
    for label, instances in sorted(label_to_instances.items()):
        counts_to_labels[len(instances)].append(label)

    sorted_labels: List[str] = []
    # Sort the labels by the number of Instances that belong to them
    for count in sorted(counts_to_labels, reverse=True):
        labels: List[str] = counts_to_labels[count]
        # Break ties by randomly shuffling labels that have the same number of Instances
        rng.shuffle(labels)
        sorted_labels.extend(labels)

    labels_iterable = cycle(sorted_labels)
    while num_instances_to_sample > 0:
        next_label: Optional[str] = next(labels_iterable, None)
        if not next_label:
            break

        instances = label_to_instances[next_label]
        # If there are no Instances to sample for this particular label, skip it.
        if len(instances) == 0:
            continue

        # Randomly sample without replacement
        examples.append(instances.pop(rng.randrange(len(instances))))
        num_instances_to_sample -= 1

    return examples



import hashlib
import json
import os
from abc import ABC, abstractmethod
from modelgauge.general import normalize_filename
from modelgauge.typed_data import Typeable, TypedData, is_typeable
from pydantic import BaseModel
from sqlitedict import SqliteDict  # type: ignore


class Cache(ABC):
    """Interface for caching."""

    @abstractmethod
    def __enter__(self):
        pass

    @abstractmethod
    def __exit__(self, *exc_info):
        pass

    @abstractmethod
    def get_or_call(self, request, callable):
        pass

    @abstractmethod
    def get_cached_response(self, request):
        pass

    @abstractmethod
    def update_cache(self, request, response):
        pass


class CacheEntry(BaseModel):
    """Wrapper around the data we write to the cache."""

    payload: TypedData


class SqlDictCache(Cache):
    """Cache the response from a method using the request as the key.

    Will create a `file_identifier`_cache.sqlite file in `data_dir` to persist
    the cache.
    """

    _CACHE_SCHEMA_VERSION = "v1"
    """Version is encoded in the table name to identify the schema."""

    def __init__(self, data_dir, file_identifier):
        os.makedirs(data_dir, exist_ok=True)
        fname = normalize_filename(f"{file_identifier}_cache.sqlite")
        path = os.path.join(data_dir, fname)
        self.cached_responses = SqliteDict(
            path,
            tablename=self._CACHE_SCHEMA_VERSION,
            encode=json.dumps,
            decode=json.loads,
        )
        tables = SqliteDict.get_tablenames(path)
        assert tables == [self._CACHE_SCHEMA_VERSION], (
            f"Expected only table to be {self._CACHE_SCHEMA_VERSION}, "
            f"but found {tables} in {path}."
        )

    def __enter__(self):
        self.cached_responses.__enter__()
        return self

    def __exit__(self, *exc_info):
        self.cached_responses.close()

    def get_or_call(self, request, callable):
        """Return the cached value, otherwise cache calling `callable`"""
        response = self.get_cached_response(request)
        if response is not None:
            return response
        response = callable(request)
        self.update_cache(request, response)
        return response

    def get_cached_response(self, request):
        """Return the cached value, or None if `request` is not in the cache."""
        if not self._can_encode(request):
            return None
        cache_key = self._hash_request(request)
        encoded_response = self.cached_responses.get(cache_key)
        if encoded_response:
            return self._decode_response(encoded_response)
        else:
            return None

    def update_cache(self, request, response: Typeable):
        """Save `response` in the cache, keyed by `request`."""
        if not self._can_encode(request) or not self._can_encode(response):
            return
        cache_key = self._hash_request(request)
        encoded_response = self._encode_response(response)
        self.cached_responses[cache_key] = encoded_response
        self.cached_responses.commit()

    def _can_encode(self, obj) -> bool:
        # Encoding currently requires Pydanic objects.
        return is_typeable(obj)

    def _encode_response(self, response: Typeable) -> str:
        return CacheEntry(payload=TypedData.from_instance(response)).model_dump_json()

    def _decode_response(self, encoded_response: str):
        return CacheEntry.model_validate_json(encoded_response).payload.to_instance()

    def _hash_request(self, request) -> str:
        return hashlib.sha256(
            TypedData.from_instance(request).model_dump_json().encode()
        ).hexdigest()


class NoCache(Cache):
    """Implements the caching interface, but never actually caches."""

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        pass

    def get_or_call(self, request, callable):
        return callable(request)

    def get_cached_response(self, request):
        return None

    def update_cache(self, request, response):
        pass



import inspect
from functools import wraps
from modelgauge.not_implemented import is_not_implemented
from modelgauge.record_init import add_initialization_record
from modelgauge.sut import SUT, PromptResponseSUT, SUTResponse
from modelgauge.sut_capabilities import (
    AcceptsChatPrompt,
    AcceptsTextPrompt,
    ProducesPerTokenLogProbabilities,
    SUTCapability,
)
from typing import Sequence, Type


def modelgauge_sut(capabilities: Sequence[Type[SUTCapability]]):
    """Decorator providing common behavior and hooks for all ModelGauge SUTs."""

    def inner(cls):
        assert issubclass(
            cls, SUT
        ), "Decorator can only be applied to classes that inherit from SUT."
        cls.capabilities = capabilities
        cls.__init__ = _wrap_init(cls.__init__)
        if issubclass(cls, PromptResponseSUT):
            _assert_prompt_types(cls)
            _override_translate_response(cls)
        cls._modelgauge_sut = True
        return cls

    return inner


def assert_is_sut(obj):
    """Raise AssertionError if obj is not decorated with @modelgauge_sut."""
    if not getattr(obj, "_modelgauge_sut", False):
        raise AssertionError(
            f"{obj.__class__.__name__} should be decorated with @modelgauge_sut."
        )


def _wrap_init(init):
    """Wrap the SUT __init__ function to verify it behaves as expected."""

    if hasattr(init, "_modelgauge_wrapped"):
        # Already wrapped, no need to do any work.
        return init

    _validate_init_signature(init)

    @wraps(init)
    def wrapped_init(self, *args, **kwargs):
        init(self, *args, **kwargs)
        add_initialization_record(self, *args, **kwargs)

    wrapped_init._modelgauge_wrapped = True
    return wrapped_init


def _validate_init_signature(init):
    params = list(inspect.signature(init).parameters.values())
    assert params[1].name == "uid", "All SUTs must have UID as the first parameter."


def _override_translate_response(cls: Type[PromptResponseSUT]) -> None:
    """Wrap the SUT translate_response function to verify it behaves as expected."""

    original = cls.translate_response

    if hasattr(original, "_modelgauge_wrapped"):
        # Already wrapped, no need to do any work.
        return

    @wraps(original)
    def inner(self, request, response) -> SUTResponse:
        response = original(self, request, response)
        logprob_capable = ProducesPerTokenLogProbabilities in self.capabilities
        logprob_produced = False
        for completion in response.completions:
            logprob_produced |= completion.top_logprobs is not None
        if not logprob_capable and logprob_produced:
            raise AssertionError(
                f"{self.__class__.__name__} does not list capability "
                f"ProducesPerTokenLogProbabilities, but it sets the top_logprobs field."
            )
        # We can't assert the other way, as if the SUTOption isn't set, the SUT may
        # not return top_logprobs.
        return response

    inner._modelgauge_wrapped = True  # type: ignore [attr-defined]
    cls.translate_response = inner  # type: ignore [method-assign]


def _assert_prompt_types(cls: Type[PromptResponseSUT]):
    _assert_prompt_type(cls, AcceptsTextPrompt, cls.translate_text_prompt)
    _assert_prompt_type(cls, AcceptsChatPrompt, cls.translate_chat_prompt)


def _assert_prompt_type(cls, capability, method):
    accepts_type = capability in cls.capabilities
    implements_type = not is_not_implemented(method)
    if accepts_type and not implements_type:
        raise AssertionError(
            f"{cls.__name__} says it {capability.__name__}, but it does not implement {method.__name__}."
        )
    if not accepts_type and implements_type:
        raise AssertionError(
            f"{cls.__name__} implements {method.__name__}, but it does not say it {capability.__name__}."
        )



from modelgauge.base_test import BaseTest
from modelgauge.sut import SUT
from modelgauge.sut_capabilities import SUTCapability
from typing import Sequence, Type


def assert_sut_capabilities(sut: SUT, test: BaseTest):
    """Raise a MissingSUTCapabilities if `sut` can't handle `test."""
    missing = []
    for capability in test.requires_sut_capabilities:
        if capability not in sut.capabilities:
            missing.append(capability)
    if missing:
        raise MissingSUTCapabilities(
            sut_uid=sut.uid, test_uid=test.uid, missing=missing
        )


def sut_is_capable(test: BaseTest, sut: SUT) -> bool:
    """Return True if `sut` can handle `test`."""
    try:
        assert_sut_capabilities(sut, test)
        return True
    except MissingSUTCapabilities:
        return False


def get_capable_suts(test: BaseTest, suts: Sequence[SUT]) -> Sequence[SUT]:
    """Filter `suts` to only those that can do `test`."""
    return [sut for sut in suts if sut_is_capable(test, sut)]


class MissingSUTCapabilities(AssertionError):
    def __init__(
        self, sut_uid: str, test_uid: str, missing: Sequence[Type[SUTCapability]]
    ):
        self.sut_uid = sut_uid
        self.test_uid = test_uid
        self.missing = missing

    def __str__(self):
        missing_names = [m.__name__ for m in self.missing]
        return (
            f"Test {self.test_uid} cannot run on {self.sut_uid} because "
            f"it requires the following capabilities: {missing_names}."
        )



import os
import tempfile
import zstandard
from abc import ABC, abstractmethod
from modelgauge.general import shell


class DataDecompressor(ABC):
    """Base class for a method which decompresses a single file into a single file."""

    @abstractmethod
    def decompress(self, compressed_location, desired_decompressed_filename: str):
        pass


class GzipDecompressor(DataDecompressor):
    def decompress(self, compressed_location: str, desired_decompressed_filename: str):
        with tempfile.TemporaryDirectory() as tmpdirname:
            # Copy file to a temp directory to not pollute original directory.
            unzipped_path = os.path.join(tmpdirname, "tmp")
            gzip_path = unzipped_path + ".gz"
            shell(["cp", compressed_location, gzip_path])
            # gzip writes its output to a file named the same as the input file, omitting the .gz extension.
            shell(["gzip", "-d", gzip_path])
            shell(["mv", unzipped_path, desired_decompressed_filename])


class ZstdDecompressor(DataDecompressor):
    def decompress(self, compressed_location: str, desired_decompressed_filename: str):
        dctx = zstandard.ZstdDecompressor()
        with open(compressed_location, "rb") as ifh:
            with open(desired_decompressed_filename, "wb") as ofh:
                dctx.copy_stream(ifh, ofh)


class DataUnpacker(ABC):
    """Base class for a method that converts a single file into a directory."""

    @abstractmethod
    def unpack(self, packed_location: str, desired_unpacked_dir: str):
        pass


class TarPacker(DataUnpacker):
    def unpack(self, packed_location: str, desired_unpacked_dir: str):
        shell(["tar", "xf", packed_location, "-C", desired_unpacked_dir])


class ZipPacker(DataUnpacker):
    def unpack(self, packed_location: str, desired_unpacked_dir: str):
        shell(["unzip", packed_location, "-d", desired_unpacked_dir])



import os
import shutil
import tomli
from importlib import resources
from modelgauge import config_templates
from modelgauge.secret_values import MissingSecretValues, RawSecrets, SecretDescription
from typing import Dict, Mapping, Sequence

DEFAULT_CONFIG_DIR = "config"
DEFAULT_SECRETS = "secrets.toml"
SECRETS_PATH = os.path.join(DEFAULT_CONFIG_DIR, DEFAULT_SECRETS)
CONFIG_TEMPLATES = [DEFAULT_SECRETS]


def write_default_config(dir: str = DEFAULT_CONFIG_DIR):
    """If the config directory doesn't exist, fill it with defaults."""
    if os.path.exists(dir):
        # Assume if it exists we don't need to add templates
        return
    os.makedirs(dir)
    for template in CONFIG_TEMPLATES:
        source_file = str(resources.files(config_templates) / template)
        output_file = os.path.join(dir, template)
        shutil.copyfile(source_file, output_file)


def load_secrets_from_config(path: str = SECRETS_PATH) -> RawSecrets:
    """Load the toml file and verify it is shaped as expected."""
    with open(path, "rb") as f:
        data = tomli.load(f)
    for values in data.values():
        # Verify the config is shaped as expected.
        assert isinstance(values, Mapping), "All keys should be in a [scope]."
        for key, value in values.items():
            assert isinstance(key, str)
            assert isinstance(value, str)
    return data


def toml_format_secrets(secrets: Sequence[SecretDescription]) -> str:
    """Format the secrets as they'd appear in a toml file.

    All values are set to "<value>".
    """

    scopes: Dict[str, Dict[str, str]] = {}
    for secret in secrets:
        if secret.scope not in scopes:
            scopes[secret.scope] = {}
        scopes[secret.scope][secret.key] = secret.instructions
    scope_displays = []
    for scope, in_scope in sorted(scopes.items()):
        scope_display = f"[{scope}]\n"
        for key, instruction in sorted(in_scope.items()):
            scope_display += f"# {instruction}\n"
            scope_display += f'{key}="<value>"\n'
        scope_displays.append(scope_display)
    return "\n".join(scope_displays)


class MissingSecretsFromConfig(MissingSecretValues):
    """Exception showing how to add missing secrets to the config file."""

    def __init__(self, missing: MissingSecretValues, config_path: str = SECRETS_PATH):
        super().__init__(descriptions=missing.descriptions)
        self.config_path = config_path

    def __str__(self):
        message = f"To perform this run you need to add the following values "
        message += f"to your secrets file '{self.config_path}':\n"
        message += toml_format_secrets(self.descriptions)
        return message


def raise_if_missing_from_config(
    missing_values: Sequence[MissingSecretValues], config_path: str = SECRETS_PATH
):
    """If there are missing secrets, raise a MissingSecretsFromConfig exception."""
    if not missing_values:
        return
    combined = MissingSecretValues.combine(missing_values)
    raise MissingSecretsFromConfig(combined, config_path)



from abc import ABC, abstractmethod


class SUTCapability(ABC):
    """Base class for defining a capability that SUTs may have and Tests may need."""

    @classmethod
    @abstractmethod
    def description(cls) -> str:
        """Describe why to mark a SUT/Test as having/needing this capability."""
        pass


class AcceptsTextPrompt(SUTCapability):
    @classmethod
    def description(cls) -> str:
        return "These SUTs can take a `TextPrompt` as input."


class AcceptsChatPrompt(SUTCapability):
    @classmethod
    def description(cls) -> str:
        return "These SUTs can take a `ChatPrompt` as input."


class ProducesPerTokenLogProbabilities(SUTCapability):
    @classmethod
    def description(cls) -> str:
        return "These SUTs set the 'top_logprobs' field in SUTResponse."



import inspect
from dataclasses import dataclass
from functools import wraps
from modelgauge.base_test import BaseTest, PromptResponseTest
from modelgauge.dependency_helper import DependencyHelper
from modelgauge.prompt import ChatPrompt, TextPrompt
from modelgauge.record_init import add_initialization_record
from modelgauge.single_turn_prompt_response import TestItem
from modelgauge.sut_capabilities import (
    AcceptsChatPrompt,
    AcceptsTextPrompt,
    ProducesPerTokenLogProbabilities,
    SUTCapability,
)
from typing import List, Sequence, Type


def modelgauge_test(requires_sut_capabilities: Sequence[Type[SUTCapability]]):
    """Decorator providing common behavior and hooks for all ModelGauge Tests."""

    def inner(cls):
        assert issubclass(
            cls, BaseTest
        ), "Decorator can only be applied to classes that inherit from BaseTest."
        cls.requires_sut_capabilities = requires_sut_capabilities
        cls.__init__ = _wrap_init(cls.__init__)
        if issubclass(cls, PromptResponseTest):
            _override_make_test_items(cls)
        cls._modelgauge_test = True
        return cls

    return inner


def assert_is_test(obj):
    """Raise AssertionError if obj is not decorated with @modelgauge_test."""
    if not getattr(obj, "_modelgauge_test", False):
        raise AssertionError(
            f"{obj.__class__.__name__} should be decorated with @modelgauge_test."
        )


def _wrap_init(init):
    """Wrap the Test __init__ function to verify it behaves as expected."""

    if hasattr(init, "_modelgauge_wrapped"):
        # Already wrapped, no need to do any work.
        return init

    _validate_init_signature(init)

    @wraps(init)
    def wrapped_init(self, *args, **kwargs):
        init(self, *args, **kwargs)
        add_initialization_record(self, *args, **kwargs)

    wrapped_init._modelgauge_wrapped = True
    return wrapped_init


def _validate_init_signature(init):
    params = list(inspect.signature(init).parameters.values())
    assert params[1].name == "uid", "All Tests must have UID as the first parameter."


@dataclass
class PromptTypeHandling:
    """Helper class for verifying the handling of a prompt type."""

    prompt_type: Type
    capability: Type[SUTCapability]
    test_obj: PromptResponseTest
    produces: bool = False

    def update_producing(self, prompt):
        self.produces |= isinstance(prompt, self.prompt_type)

    def assert_handled(self):
        required = self.capability in self.test_obj.requires_sut_capabilities
        test_name = self.test_obj.__class__.__name__
        prompt_type_name = self.prompt_type.__name__
        capability_name = self.capability.__name__
        if self.produces and not required:
            raise AssertionError(
                f"{test_name} produces {prompt_type_name} but does not requires_sut_capabilities {capability_name}."
            )
        # Tests may conditionally produce a prompt type, so requirements are a superset.


def _override_make_test_items(cls: Type[PromptResponseTest]) -> None:
    """Wrap the Test make_test_items function to verify it behaves as expected."""

    original = cls.make_test_items

    if hasattr(original, "_modelgauge_wrapped"):
        # Already wrapped, no need to do any work.
        return

    @wraps(original)
    def inner(self, dependency_helper: DependencyHelper) -> List[TestItem]:
        items: List[TestItem] = original(self, dependency_helper)
        requires_logprobs = (
            ProducesPerTokenLogProbabilities in self.requires_sut_capabilities
        )
        prompt_types = [
            PromptTypeHandling(
                prompt_type=TextPrompt,
                capability=AcceptsTextPrompt,
                test_obj=self,
            ),
            PromptTypeHandling(
                prompt_type=ChatPrompt,
                capability=AcceptsChatPrompt,
                test_obj=self,
            ),
        ]
        any_request_logprobs = False
        for item in items:
            for prompt in item.prompts:
                any_request_logprobs |= prompt.prompt.options.top_logprobs is not None
                for prompt_type in prompt_types:
                    prompt_type.update_producing(prompt.prompt)

        if any_request_logprobs and not requires_logprobs:
            raise AssertionError(
                f"{self.__class__.__name__} specified the SUT option top_logprobs, "
                f"but did not list ProducesPerTokenLogProbabilities as a "
                f"required capability. If it doesn't actually need top_logprobs, "
                f"remove setting the option."
            )

        if not any_request_logprobs and requires_logprobs:
            raise AssertionError(
                f"{self.__class__.__name__} lists ProducesPerTokenLogProbabilities "
                f"as required, but did not request the SUT option top_logprobs. "
                f"If it doesn't actually need top_logprobs, remove specifying the capability."
            )

        for prompt_type in prompt_types:
            prompt_type.assert_handled()
        return items

    inner._modelgauge_wrapped = True  # type: ignore [attr-defined]
    cls.make_test_items = inner  # type: ignore [method-assign]



import os
import click
from modelgauge.base_test import PromptResponseTest
from modelgauge.command_line import (
    DATA_DIR_OPTION,
    MAX_TEST_ITEMS_OPTION,
    SUT_OPTION,
    display_header,
    display_list_item,
    modelgauge_cli,
)
from modelgauge.config import (
    load_secrets_from_config,
    raise_if_missing_from_config,
    toml_format_secrets,
)
from modelgauge.dependency_injection import list_dependency_usage
from modelgauge.general import normalize_filename
from modelgauge.instance_factory import FactoryEntry
from modelgauge.load_plugins import list_plugins, load_plugins
from modelgauge.prompt import SUTOptions, TextPrompt
from modelgauge.simple_test_runner import run_prompt_response_test
from modelgauge.secret_values import MissingSecretValues, RawSecrets, get_all_secrets
from modelgauge.sut import PromptResponseSUT
from modelgauge.sut_registry import SUTS
from modelgauge.test_registry import TESTS
from typing import List, Optional


@modelgauge_cli.command(name="list")
def list_command() -> None:
    """Overview of Plugins, Tests, and SUTs."""
    plugins = list_plugins()
    display_header(f"Plugin Modules: {len(plugins)}")
    for module_name in plugins:
        display_list_item(module_name)
    suts = SUTS.items()
    display_header(f"SUTS: {len(suts)}")
    for sut, sut_entry in suts:
        display_list_item(f"{sut} {sut_entry}")
    tests = TESTS.items()
    display_header(f"Tests: {len(tests)}")
    for test, test_entry in tests:
        display_list_item(f"{test} {test_entry}")


def _display_factory_entry(uid: str, entry: FactoryEntry, secrets: RawSecrets):
    def format_missing_secrets(missing):
        """Return formatted string for missing secrets."""
        return "\n".join(
            f"Scope: '{secret['scope']}', Key: '{secret['key']}', Instructions: '{secret['instructions']}'"
            for secret in missing
        )

    used, missing = list_dependency_usage(entry.args, entry.kwargs, secrets)
    missing = format_missing_secrets(missing)

    display_header(uid)
    click.echo(f"Class: {entry.cls.__name__}")
    click.echo(f"Args: {entry.args}")
    click.echo(f"Kwargs: {entry.kwargs}")

    if used:
        click.echo("Used Secrets:")
        click.echo(used)
    else:
        click.echo("No Secrets Used.")

    if missing:
        click.echo("Missing Secrets:")
        click.echo(missing)

    click.echo()


@modelgauge_cli.command()
def list_tests() -> None:
    """List details about all registered tests."""
    secrets = load_secrets_from_config()
    for test_uid, test_entry in TESTS.items():
        _display_factory_entry(test_uid, test_entry, secrets)


@modelgauge_cli.command()
def list_suts():
    """List details about all registered SUTs (System Under Test)."""
    secrets = load_secrets_from_config()
    for sut_uid, sut in SUTS.items():
        _display_factory_entry(sut_uid, sut, secrets)


@modelgauge_cli.command()
def list_secrets() -> None:
    """List details about secrets modelgauge might need."""
    descriptions = get_all_secrets()
    if descriptions:
        display_header("Here are the known secrets modelgauge might use.")
        click.echo(toml_format_secrets(descriptions))
    else:
        display_header("No secrets used by any installed plugin.")


@modelgauge_cli.command()
@SUT_OPTION
@click.option("--prompt", help="The full text to send to the SUT.")
@click.option(
    "--num-completions",
    default=None,
    type=click.IntRange(1),
    help="How many different completions to generation.",
)
@click.option(
    "--max-tokens",
    default=None,
    type=click.IntRange(1),
    help="How many tokens to generate for each completion.",
)
@click.option(
    "--top-logprobs",
    type=click.IntRange(1),
    help="How many log probabilities to report for each token position.",
)
def run_sut(
    sut: str,
    prompt: str,
    num_completions: Optional[int],
    max_tokens: Optional[int],
    top_logprobs: Optional[int],
):
    """Send a prompt from the command line to a SUT."""
    secrets = load_secrets_from_config()
    try:
        sut_obj = SUTS.make_instance(sut, secrets=secrets)
    except MissingSecretValues as e:
        raise_if_missing_from_config([e])

    # Current this only knows how to do prompt response, so assert that is what we have.
    assert isinstance(sut_obj, PromptResponseSUT)

    options = SUTOptions()
    if num_completions:
        options.num_completions = num_completions
    if max_tokens:
        options.max_tokens = max_tokens
    if top_logprobs:
        options.top_logprobs = top_logprobs
    prompt_obj = TextPrompt(text=prompt, options=options)
    request = sut_obj.translate_text_prompt(prompt_obj)
    click.echo(f"Native request: {request}\n")
    response = sut_obj.evaluate(request)
    click.echo(f"Native response: {response}\n")
    result = sut_obj.translate_response(request, response)
    click.echo(f"Normalized response: {result.model_dump_json(indent=2)}\n")


@modelgauge_cli.command()
@click.option("--test", help="Which registered TEST to run.", required=True)
@SUT_OPTION
@DATA_DIR_OPTION
@MAX_TEST_ITEMS_OPTION
@click.option(
    "--output-file",
    help="If specified, will override the default location for outputting the TestRecord.",
)
@click.option(
    "--no-caching",
    is_flag=True,
    show_default=True,
    default=False,
    help="Disable caching.",
)
@click.option(
    "--no-progress-bar",
    is_flag=True,
    show_default=True,
    default=False,
    help="Disable displaying the 'Processing TestItems' progress bar.",
)
def run_test(
    test: str,
    sut: str,
    data_dir: str,
    max_test_items: int,
    output_file: Optional[str],
    no_caching: bool,
    no_progress_bar: bool,
):
    """Run the Test on the desired SUT and output the TestRecord."""
    secrets = load_secrets_from_config()
    # Check for missing secrets without instantiating any objects
    missing_secrets: List[MissingSecretValues] = []
    missing_secrets.extend(TESTS.get_missing_dependencies(test, secrets=secrets))
    missing_secrets.extend(SUTS.get_missing_dependencies(sut, secrets=secrets))
    raise_if_missing_from_config(missing_secrets)

    test_obj = TESTS.make_instance(test, secrets=secrets)
    sut_obj = SUTS.make_instance(sut, secrets=secrets)

    # Current this only knows how to do prompt response, so assert that is what we have.
    assert isinstance(sut_obj, PromptResponseSUT)
    assert isinstance(test_obj, PromptResponseTest)

    if output_file is None:
        os.makedirs("output", exist_ok=True)
        output_file = os.path.join(
            "output", normalize_filename(f"record_for_{test}_{sut}.json")
        )
    test_record = run_prompt_response_test(
        test_obj,
        sut_obj,
        data_dir,
        max_test_items,
        use_caching=not no_caching,
        disable_progress_bar=no_progress_bar,
    )
    with open(output_file, "w") as f:
        print(test_record.model_dump_json(indent=4), file=f)
    # For displaying to the screen, clear out the verbose test_item_records
    test_record.test_item_records = []
    print(test_record.model_dump_json(indent=4))
    print("Full TestRecord json written to", output_file)


def main():
    # We need to call `load_plugins` before the cli in order to:
    # * Allow plugins to add their own CLI commands
    # * Enable --help to correctly list command options (e.g. valid values for SUT)
    load_plugins()
    modelgauge_cli()


if __name__ == "__main__":
    main()



import inspect
import threading
from dataclasses import dataclass
from modelgauge.dependency_injection import inject_dependencies
from modelgauge.secret_values import MissingSecretValues, RawSecrets
from modelgauge.tracked_object import TrackedObject
from typing import Any, Dict, Generic, List, Sequence, Tuple, Type, TypeVar

_T = TypeVar("_T", bound=TrackedObject)


@dataclass(frozen=True)
class FactoryEntry(Generic[_T]):
    """Container for how to initialize an object."""

    cls: Type[_T]
    uid: str
    args: Tuple[Any]
    kwargs: Dict[str, Any]

    def __post_init__(self):
        param_names = list(inspect.signature(self.cls).parameters.keys())
        if not param_names or param_names[0] != "uid":
            raise AssertionError(
                f"Cannot create factory entry for {self.cls} as its first "
                f"constructor argument must be 'uid'. Arguments: {param_names}."
            )

    def __str__(self):
        """Return a string representation of the entry."""
        return f"{self.cls.__name__}(uid={self.uid}, args={self.args}, kwargs={self.kwargs})"

    def make_instance(self, *, secrets: RawSecrets) -> _T:
        """Construct an instance of this object, with dependency injection."""
        args, kwargs = inject_dependencies(self.args, self.kwargs, secrets=secrets)
        result = self.cls(self.uid, *args, **kwargs)  # type: ignore [call-arg]
        assert hasattr(
            result, "uid"
        ), f"Class {self.cls} must set member variable 'uid'."
        assert (
            result.uid == self.uid
        ), f"Class {self.cls} must set 'uid' to first constructor argument."
        return result

    def get_missing_dependencies(
        self, *, secrets: RawSecrets
    ) -> Sequence[MissingSecretValues]:
        """Find all missing dependencies for this object."""
        # TODO: Handle more kinds of dependency failure.
        try:
            inject_dependencies(self.args, self.kwargs, secrets=secrets)
        except MissingSecretValues as e:
            return [e]
        return []


class InstanceFactory(Generic[_T]):
    """Generic class that lets you store how to create instances of a given type."""

    def __init__(self) -> None:
        self._lookup: Dict[str, FactoryEntry[_T]] = {}
        self.lock = threading.Lock()

    def register(self, cls: Type[_T], uid: str, *args, **kwargs):
        """Add value to the registry, ensuring it has a unique key."""

        with self.lock:
            previous = self._lookup.get(uid)
            assert previous is None, (
                f"Factory already contains {uid} set to "
                f"{previous.cls.__name__}(args={previous.args}, "
                f"kwargs={previous.kwargs})."
            )
            self._lookup[uid] = FactoryEntry[_T](cls, uid, args, kwargs)

    def make_instance(self, uid: str, *, secrets: RawSecrets) -> _T:
        """Create an instance using the  class and arguments passed to register, raise exception if missing."""
        entry = self._get_entry(uid)
        return entry.make_instance(secrets=secrets)

    def get_missing_dependencies(
        self, uid: str, *, secrets: RawSecrets
    ) -> Sequence[MissingSecretValues]:
        """Find all missing dependencies for `uid`."""
        entry = self._get_entry(uid)
        return entry.get_missing_dependencies(secrets=secrets)

    def _get_entry(self, uid: str) -> FactoryEntry:
        with self.lock:
            entry: FactoryEntry
            try:
                entry = self._lookup[uid]
            except KeyError:
                known_uids = list(self._lookup.keys())
                raise KeyError(f"No registration for {uid}. Known uids: {known_uids}")
        return entry

    def items(self) -> List[Tuple[str, FactoryEntry[_T]]]:
        """List all items in the registry."""
        with self.lock:
            return list(self._lookup.items())



import importlib
from modelgauge.dependency_injection import (
    inject_dependencies,
    serialize_injected_dependencies,
)
from modelgauge.secret_values import RawSecrets
from pydantic import BaseModel
from typing import Any, List, Mapping


class InitializationRecord(BaseModel):
    """Holds data sufficient to reconstruct an object."""

    module: str
    class_name: str
    args: List[Any]
    kwargs: Mapping[str, Any]

    def recreate_object(self, *, secrets: RawSecrets = {}):
        """Redoes the init call from this record."""
        cls = getattr(importlib.import_module(self.module), self.class_name)
        args, kwargs = inject_dependencies(self.args, self.kwargs, secrets=secrets)
        return cls(*args, **kwargs)


def add_initialization_record(self, *args, **kwargs):
    record_args, record_kwargs = serialize_injected_dependencies(args, kwargs)
    self.initialization_record = InitializationRecord(
        module=self.__class__.__module__,
        class_name=self.__class__.__qualname__,
        args=record_args,
        kwargs=record_kwargs,
    )



import gdown  # type: ignore
import shutil
import tempfile
import urllib.request
from abc import ABC, abstractmethod
from dataclasses import dataclass
from modelgauge.data_packing import DataDecompressor, DataUnpacker
from modelgauge.general import UrlRetrieveProgressBar
from typing import Optional


@dataclass(frozen=True, kw_only=True)
class ExternalData(ABC):
    """Base class for defining a source of external data."""

    decompressor: Optional[DataDecompressor] = None
    unpacker: Optional[DataUnpacker] = None

    @abstractmethod
    def download(self, location):
        pass


@dataclass(frozen=True, kw_only=True)
class WebData(ExternalData):
    """External data that can be trivially downloaded using wget."""

    source_url: str

    def download(self, location):
        urllib.request.urlretrieve(
            self.source_url,
            location,
            reporthook=UrlRetrieveProgressBar(self.source_url),
        )


@dataclass(frozen=True, kw_only=True)
class GDriveData(ExternalData):
    """File downloaded using a google drive folder url and a file's relative path to the folder."""

    data_source: str
    file_path: str

    def download(self, location):
        with tempfile.TemporaryDirectory() as tmpdir:
            # Empty folder downloaded to tmpdir
            available_files = gdown.download_folder(
                url=self.data_source, skip_download=True, quiet=True, output=tmpdir
            )
        # Find file id needed to download the file.
        for file in available_files:
            if file.path == self.file_path:
                gdown.download(id=file.id, output=location)
                return
        raise RuntimeError(
            f"Cannot find file with name {self.file_path} in google drive folder {self.data_source}"
        )


@dataclass(frozen=True, kw_only=True)
class LocalData(ExternalData):
    """A file that is already on your local machine.

    WARNING: Only use this in cases where your data is not yet
    publicly available, but will be eventually.
    """

    path: str

    def download(self, location):
        shutil.copy(self.path, location)



import datetime
import hashlib
import importlib
import inspect
import logging
import shlex
import subprocess
import time
from tqdm import tqdm
from typing import Dict, List, Optional, Set, Type, TypeVar

# Type vars helpful in defining templates.
_InT = TypeVar("_InT")


def current_timestamp_millis() -> int:
    return time.time_ns() // 1_000_000


def get_concrete_subclasses(cls: Type[_InT]) -> Set[Type[_InT]]:
    result = set()
    for subclass in cls.__subclasses__():
        if not inspect.isabstract(subclass):
            result.add(subclass)
        result.update(get_concrete_subclasses(subclass))
    return result


def value_or_default(value: Optional[_InT], default: _InT) -> _InT:
    if value is not None:
        return value
    return default


def shell(args: List[str]):
    """Executes the shell command in `args`."""
    cmd = shlex.join(args)
    logging.info(f"Executing: {cmd}")
    exit_code = subprocess.call(args)
    if exit_code != 0:
        logging.error(f"Failed with exit code {exit_code}: {cmd}")


def hash_file(filename, block_size=65536):
    """Apply sha256 to the bytes of `filename`."""
    file_hash = hashlib.sha256()
    with open(filename, "rb") as f:
        while True:
            block = f.read(block_size)
            if not block:
                break
            file_hash.update(block)

    return file_hash.hexdigest()


def normalize_filename(filename: str) -> str:
    """Replace filesystem characters in `filename`."""
    return filename.replace("/", "_")


class UrlRetrieveProgressBar:
    """Progress bar compatible with urllib.request.urlretrieve."""

    def __init__(self, url: str):
        self.bar = None
        self.url = url

    def __call__(self, block_num, block_size, total_size):
        if not self.bar:
            self.bar = tqdm(total=total_size, unit="B", unit_scale=True)
            self.bar.set_description(f"Downloading {self.url}")
        self.bar.update(block_size)


def get_class(module_name: str, qual_name: str):
    """Get the class object given its __module__ and __qualname__."""
    scope = importlib.import_module(module_name)
    names = qual_name.split(".")
    for name in names:
        scope = getattr(scope, name)
    return scope


def current_local_datetime():
    """Get the current local date time, with timezone."""
    return datetime.datetime.now().astimezone()



import math
from collections import defaultdict
from modelgauge.single_turn_prompt_response import MeasuredTestItem
from pydantic import BaseModel
from typing import Callable, List, Mapping, Sequence, TypeVar


def get_measurements(
    measurement_name: str, items: List[MeasuredTestItem]
) -> List[float]:
    """Extract a desired measurement for all TestItems."""
    # Raises a KeyError if that test item is missing that measurement.
    return [item.measurements[measurement_name] for item in items]


class MeasurementStats(BaseModel):
    """Container for common stats about a measurement."""

    sum: float
    mean: float
    count: int
    population_variance: float
    population_std_dev: float
    # TODO Consider min, max, and median

    @staticmethod
    def calculate(values: Sequence[float]) -> "MeasurementStats":
        if len(values) == 0:
            return MeasurementStats(
                sum=0, mean=0, count=0, population_variance=0, population_std_dev=0
            )
        total = sum(values)
        count = len(values)
        mean = total / count
        deviations = [(x - mean) ** 2 for x in values]
        variance = sum(deviations) / len(values)
        std_dev = math.sqrt(variance)
        return MeasurementStats(
            sum=total,
            mean=mean,
            count=count,
            population_variance=variance,
            population_std_dev=std_dev,
        )


def get_measurement_stats(
    measurement_name: str, items: List[MeasuredTestItem]
) -> MeasurementStats:
    """Calculate common statistics about `measurement_name`."""
    values = get_measurements(measurement_name, items)
    return MeasurementStats.calculate(values)


_T = TypeVar("_T")


def get_measurement_stats_by_key(
    measurement_name: str,
    items: List[MeasuredTestItem],
    *,
    key: Callable[[MeasuredTestItem], _T]
) -> Mapping[_T, MeasurementStats]:
    """Calculate statistics grouping by `key`."""
    groups = defaultdict(list)
    for item in items:
        groups[key(item)].append(item)
    stats = {}
    for key_value, measurements in groups.items():
        stats[key_value] = get_measurement_stats(measurement_name, measurements)
    return stats


def sum_measurements(measurement_name: str, items: List[MeasuredTestItem]) -> float:
    measurements = get_measurements(measurement_name, items)
    return sum(measurement for measurement in measurements)


def mean_of_measurement(measurement_name: str, items: List[MeasuredTestItem]) -> float:
    """Calculate the mean across all TestItems for a desired measurement."""
    measurements = get_measurements(measurement_name, items)
    total = sum(measurements)
    return total / len(measurements)



from modelgauge.base_test import TestResult
from modelgauge.general import current_local_datetime
from modelgauge.record_init import InitializationRecord
from modelgauge.single_turn_prompt_response import (
    PromptInteractionAnnotations,
    TestItem,
)
from pydantic import AwareDatetime, BaseModel, Field
from typing import Dict, List, Mapping


class TestItemRecord(BaseModel):
    """Record of all data relevant to a single TestItem."""

    # TODO: This duplicates the list of prompts across test_item and interactions.
    # Maybe just copy the TestItem context.
    test_item: TestItem
    interactions: List[PromptInteractionAnnotations]
    measurements: Dict[str, float]

    __test__ = False


class TestRecord(BaseModel):
    """Record of all data relevant to a single run of a Test."""

    run_timestamp: AwareDatetime = Field(default_factory=current_local_datetime)
    test_uid: str
    test_initialization: InitializationRecord
    dependency_versions: Mapping[str, str]
    sut_uid: str
    sut_initialization: InitializationRecord
    # TODO We should either reintroduce "Turns" here, or expect
    # there to b different schemas for different TestImplementationClasses.
    test_item_records: List[TestItemRecord]
    result: TestResult

    __test__ = False



import os
import random
from modelgauge.annotation import Annotation
from modelgauge.annotator import CompletionAnnotator
from modelgauge.base_test import PromptResponseTest, TestResult
from modelgauge.caching import Cache, NoCache, SqlDictCache
from modelgauge.dependency_helper import FromSourceDependencyHelper
from modelgauge.prompt import TextPrompt
from modelgauge.records import TestItemRecord, TestRecord
from modelgauge.single_turn_prompt_response import (
    MeasuredTestItem,
    PromptInteractionAnnotations,
    SUTCompletionAnnotations,
    SUTResponseAnnotations,
    TestItem,
    TestItemAnnotations,
)
from modelgauge.sut import PromptResponseSUT
from modelgauge.sut_capabilities_verification import assert_sut_capabilities
from modelgauge.sut_decorator import assert_is_sut
from modelgauge.test_decorator import assert_is_test
from tqdm import tqdm
from typing import List, Optional


def run_prompt_response_test(
    test: PromptResponseTest,
    sut: PromptResponseSUT,
    data_dir: str,
    max_test_items: Optional[int] = None,
    use_caching: bool = True,
    disable_progress_bar: bool = False,
) -> TestRecord:
    """Demonstration for how to run a single Test on a single SUT, all calls serial."""

    assert_is_test(test)
    assert_is_sut(sut)
    assert_sut_capabilities(sut, test)

    # Ensure we can record what these objects are
    test_initialization = test.initialization_record
    sut_initialization = sut.initialization_record
    test_data_path = os.path.join(data_dir, "tests", test.__class__.__name__)

    sut_cache: Cache
    if use_caching:
        sut_cache = SqlDictCache(os.path.join(data_dir, "suts"), sut.uid)
    else:
        sut_cache = NoCache()
    annotators = []
    for key, annotator in test.get_annotators().items():
        annotator_cache: Cache
        if use_caching:
            annotator_cache = SqlDictCache(
                os.path.join(test_data_path, "annotators"), key
            )
        else:
            annotator_cache = NoCache()
        assert isinstance(
            annotator, CompletionAnnotator
        ), "Only know how to do CompletionAnnotator."
        annotators.append(AnnotatorData(key, annotator, annotator_cache))

    # This runner just records versions, it doesn't specify a required version.
    dependency_helper = FromSourceDependencyHelper(
        os.path.join(test_data_path, "dependency_data"),
        test.get_dependencies(),
        required_versions={},
    )

    test_items = test.make_test_items(dependency_helper)
    if max_test_items is not None:
        assert max_test_items > 0, f"Cannot run a test using {max_test_items}."
        if max_test_items < len(test_items):
            rng = random.Random()
            rng.seed(0)
            rng.shuffle(test_items)
            test_items = test_items[:max_test_items]
    test_item_records = []
    measured_test_items = []
    desc = f"Processing TestItems for test={test.uid} sut={sut.uid}"
    for test_item in tqdm(test_items, desc=desc, disable=disable_progress_bar):
        test_item_record = _process_test_item(
            test_item, test, sut, sut_cache, annotators
        )
        test_item_records.append(test_item_record)
        measured_test_items.append(
            MeasuredTestItem(
                test_item=test_item_record.test_item,
                measurements=test_item_record.measurements,
            )
        )
    test_result = TestResult.from_instance(
        test.aggregate_measurements(measured_test_items)
    )
    return TestRecord(
        test_uid=test.uid,
        test_initialization=test_initialization,
        dependency_versions=dependency_helper.versions_used(),
        sut_uid=sut.uid,
        sut_initialization=sut_initialization,
        test_item_records=test_item_records,
        result=test_result,
    )


class AnnotatorData:
    """Container to hold data about an annotator."""

    def __init__(self, key: str, annotator: CompletionAnnotator, cache: Cache):
        self.key = key
        self.annotator = annotator
        self.cache = cache


def _process_test_item(
    item: TestItem,
    test: PromptResponseTest,
    sut: PromptResponseSUT,
    sut_cache: Cache,
    annotators: List[AnnotatorData],
) -> TestItemRecord:
    interactions: List[PromptInteractionAnnotations] = []
    for prompt in item.prompts:
        try:
            if isinstance(prompt.prompt, TextPrompt):
                sut_request = sut.translate_text_prompt(prompt.prompt)
            else:
                sut_request = sut.translate_chat_prompt(prompt.prompt)
            with sut_cache as cache:
                sut_response = cache.get_or_call(sut_request, sut.evaluate)
            response = sut.translate_response(sut_request, sut_response)
        except Exception as e:
            raise Exception(
                f"Exception while handling SUT {sut.uid} for TestItem `{item}`"
            ) from e

        annotated_completions: List[SUTCompletionAnnotations] = []
        for completion in response.completions:
            annotations = {}
            for annotator_data in annotators:
                annotator = annotator_data.annotator
                try:
                    annotator_request = annotator.translate_request(prompt, completion)
                    with annotator_data.cache as cache:
                        annotator_response = cache.get_or_call(
                            annotator_request, annotator.annotate
                        )
                    annotation = annotator.translate_response(
                        annotator_request, annotator_response
                    )
                except Exception as e:
                    raise Exception(
                        f"Exception while handling annotation for {annotator_data.key} on {response}"
                    ) from e

                annotations[annotator_data.key] = Annotation.from_instance(annotation)
            annotated_completions.append(
                SUTCompletionAnnotations(completion=completion, annotations=annotations)
            )
        interactions.append(
            PromptInteractionAnnotations(
                prompt=prompt,
                response=SUTResponseAnnotations(completions=annotated_completions),
            )
        )
    annotated = TestItemAnnotations(
        test_item=item,
        interactions=interactions,
    )
    measurements = test.measure_quality(annotated)

    return TestItemRecord(
        test_item=annotated.test_item,
        interactions=annotated.interactions,
        measurements=measurements,
    )



from enum import Enum
from pydantic import BaseModel
from typing import List, Optional


class SUTOptions(BaseModel):
    """
    An exhaustive set of options that could potentially be desired by a SUT.

    Not all SUTs respect all options.
    """

    num_completions: int = 1
    """Generate this many completions (by sampling from the model)"""

    max_tokens: int = 100
    """Maximum number of tokens to generate (per completion)"""

    temperature: Optional[float] = None
    """Temperature parameter that governs diversity"""

    top_k_per_token: Optional[int] = None
    """Take this many highest probability candidates per token in the completion"""

    stop_sequences: Optional[List[str]] = None
    """Stop generating once we hit one of these strings."""

    top_p: Optional[float] = None
    """Same from tokens that occupy this probability mass (nucleus sampling)"""

    presence_penalty: Optional[float] = None
    """Penalize repetition (OpenAI & Writer only)"""

    frequency_penalty: Optional[float] = None
    """Penalize repetition (OpenAI & Writer only)"""

    random: Optional[str] = None
    """Used to control randomness. Expect different responses for the same
    request but with different values for `random`."""

    # Must specify SUTCapabilities for these
    top_logprobs: Optional[int] = None
    """If present, will request the log probabilities for this
    many of the top tokens at each token position."""


class ChatRole(str, Enum):
    user = "USER"
    sut = "SUT"


class ChatMessage(BaseModel):
    text: str
    role: ChatRole


class ChatPrompt(BaseModel):
    messages: List[ChatMessage]
    options: SUTOptions = SUTOptions()


class TextPrompt(BaseModel):
    """What actually goes to the SUT."""

    text: str
    options: SUTOptions = SUTOptions()



from modelgauge.typed_data import TypedData


class Annotation(TypedData):
    """Container for plugin defined annotation data.

    Every annotator can return data however it wants.
    Since Tests are responsible for both deciding what
    Annotators to apply and how to interpret their results,
    they can use `to_instance` to get it back in the form they want.
    """

    pass



from abc import ABC, abstractmethod
from dataclasses import dataclass
from modelgauge.general import get_concrete_subclasses
from pydantic import BaseModel
from typing import Generic, List, Mapping, Optional, Sequence, Type, TypeVar


class SecretDescription(BaseModel):
    """How to look up a secret and how to get the value if you don't have it."""

    scope: str
    key: str
    instructions: str


RawSecrets = Mapping[str, Mapping[str, str]]
"""Convenience typing for how the secrets are read from a file."""


SecretType = TypeVar("SecretType", bound="Secret")


@dataclass(frozen=True)
class Secret(ABC):
    """Base class for all secrets."""

    @classmethod
    @abstractmethod
    def description(cls) -> SecretDescription:
        """Information about how to lookup/obtain the secret."""
        pass

    @classmethod
    @abstractmethod
    def make(cls: Type[SecretType], raw_secrets: RawSecrets) -> SecretType:
        """Read the secret value from `raw_secrets to make this class."""
        pass


def get_all_secrets() -> Sequence[SecretDescription]:
    """Return the descriptions of all possible secrets."""
    secrets = get_concrete_subclasses(Secret)  # type: ignore
    return [s.description() for s in secrets]


class SerializedSecret(BaseModel):
    """Hold a pointer to the secret class in a serializable form."""

    module: str
    class_name: str

    @staticmethod
    def serialize(secret: Secret) -> "SerializedSecret":
        """Create a SerializedSecret from a Secret"""
        return SerializedSecret(
            module=secret.__class__.__module__,
            class_name=secret.__class__.__qualname__,
        )


RequiredSecretType = TypeVar("RequiredSecretType", bound="RequiredSecret")


class RequiredSecret(Secret):
    """Base class for all required secrets."""

    def __init__(self, value: str):
        super().__init__()
        self._value = value

    @property
    def value(self) -> str:
        """Get the value of the secret."""
        return self._value

    @classmethod
    def make(
        cls: Type[RequiredSecretType], raw_secrets: RawSecrets
    ) -> RequiredSecretType:
        """Construct this class from the data in raw_secrets.

        Raises MissingSecretValues if desired secret is missing.
        """
        secret = cls.description()
        try:
            return cls(raw_secrets[secret.scope][secret.key])
        except KeyError:
            raise MissingSecretValues([secret])


class MissingSecretValues(LookupError):
    """Exception describing one or more missing required secrets."""

    def __init__(self, descriptions: Sequence[SecretDescription]):
        assert descriptions, "Must have at least 1 description to raise an error."
        self.descriptions = descriptions

    @staticmethod
    def combine(errors: Sequence["MissingSecretValues"]) -> "MissingSecretValues":
        """Combine multiple exceptions into one."""
        descriptions: List[SecretDescription] = []
        for error in errors:
            descriptions.extend(error.descriptions)
        return MissingSecretValues(descriptions)

    def __str__(self):
        message = "Missing the following secrets:\n"
        for d in self.descriptions:
            # TODO Make this nicer.
            message += str(d) + "\n"
        return message


OptionalSecretType = TypeVar("OptionalSecretType", bound="OptionalSecret")


class OptionalSecret(Secret):
    """Base class for all optional secrets."""

    def __init__(self, value: Optional[str]):
        super().__init__()
        self._value = value

    @property
    def value(self) -> Optional[str]:
        """Get the secret value, or None if it wasn't provided."""
        return self._value

    @classmethod
    def make(
        cls: Type[OptionalSecretType], raw_secrets: RawSecrets
    ) -> OptionalSecretType:
        """Construct this class from the data in raw_secrets.

        Sets value to None if desired secret is missing.
        """
        secret = cls.description()
        try:
            return cls(raw_secrets[secret.scope][secret.key])
        except KeyError:
            return cls(None)


_T = TypeVar("_T")


# TODO Consider moving these to dependency_injection.py
class Injector(ABC, Generic[_T]):
    """Base class for delayed injection of a value."""

    @abstractmethod
    def inject(self, raw_secrets: RawSecrets) -> _T:
        """Use `raw_secrets` to construct the object."""
        pass


class InjectSecret(Injector, Generic[SecretType]):
    def __init__(self, secret_class: Type[SecretType]):
        self.secret_class = secret_class

    def inject(self, raw_secrets: RawSecrets) -> SecretType:
        return self.secret_class.make(raw_secrets)

    def __repr__(self):
        return f"InjectSecret({self.secret_class.__name__})"



from abc import ABC, abstractmethod
from modelgauge.not_implemented import not_implemented
from modelgauge.prompt import ChatPrompt, TextPrompt
from modelgauge.record_init import InitializationRecord
from modelgauge.sut_capabilities import SUTCapability
from modelgauge.tracked_object import TrackedObject
from pydantic import BaseModel
from typing import Generic, List, Optional, Sequence, Type, TypeVar

RequestType = TypeVar("RequestType")
ResponseType = TypeVar("ResponseType")


class TokenProbability(BaseModel):
    """Probability assigned to a given token."""

    token: str
    logprob: float


class TopTokens(BaseModel):
    """List of most likely tokens and their probabilities."""

    top_tokens: Sequence[TokenProbability]


class SUTCompletion(BaseModel):
    """All data about a single completion in the response."""

    text: str
    top_logprobs: Optional[Sequence[TopTokens]] = None
    """For each position, list the probabilities for each of the most likely tokens.

    To guarantee this field is not None, the Test must specify SUTOptions.top_logprobs
    and that it requires_sut_capabilities ProducesPerTokenLogProbabilities.
    SUTs that set this value must specify they have the ProducesPerTokenLogProbabilities
    capability. They may conditional setting the field on on SUTOptions.top_logprobs being not None.
    """


class SUTResponse(BaseModel):
    """The data that came out of the SUT."""

    completions: List[SUTCompletion]


class SUT(TrackedObject):
    """Base class for all SUTs. There is no guaranteed interface between SUTs, so no methods here."""

    # Set automatically by @modelgauge_sut()
    capabilities: Sequence[Type[SUTCapability]]

    def __init__(self, uid: str):
        super().__init__(uid)
        # The initialization record is set automatically by @modelgauge_sut()
        self.initialization_record: InitializationRecord


class PromptResponseSUT(SUT, ABC, Generic[RequestType, ResponseType]):
    """The base class for any SUT that is designed for handling a single-turn."""

    @not_implemented
    def translate_text_prompt(self, prompt: TextPrompt) -> RequestType:
        """Convert the prompt into the SUT's native representation."""
        raise NotImplementedError(
            f"SUT {self.__class__.__name__} does not implement translate_text_prompt."
        )

    @not_implemented
    def translate_chat_prompt(self, prompt: ChatPrompt) -> RequestType:
        """Convert the prompt into the SUT's native representation."""
        raise NotImplementedError(
            f"SUT {self.__class__.__name__} does not implement translate_chat_prompt."
        )

    @abstractmethod
    def evaluate(self, request: RequestType) -> ResponseType:
        """Evaluate this SUT on the native request."""
        pass

    @abstractmethod
    def translate_response(
        self, request: RequestType, response: ResponseType
    ) -> SUTResponse:
        """Convert the native response into a form all Tests can process."""
        pass



import click
from modelgauge.config import write_default_config


@click.group()
def modelgauge_cli():
    """Run the ModelGauge library from the command line."""
    # To add a command, decorate your function with @modelgauge_cli.command().
    # Always create the config directory if it doesn't already exist.
    write_default_config()


def display_header(text):
    """Echo the text, but in bold!"""
    click.echo(click.style(text, bold=True))


def display_list_item(text):
    click.echo(f"\t{text}")


# Define some reusable options
DATA_DIR_OPTION = click.option(
    "--data-dir",
    default="run_data",
    help="Where to store the auxiliary data produced during the run.",
)

MAX_TEST_ITEMS_OPTION = click.option(
    "-m",
    "--max-test-items",
    default=None,
    type=click.IntRange(1),  # Must be a postive integer
    help="Maximum number of TestItems a Test should run.",
)

SUT_OPTION = click.option("--sut", help="Which registered SUT to run.", required=True)



from abc import ABC, abstractmethod
from modelgauge.annotator import Annotator
from modelgauge.dependency_helper import DependencyHelper
from modelgauge.external_data import ExternalData
from modelgauge.record_init import InitializationRecord
from modelgauge.single_turn_prompt_response import (
    MeasuredTestItem,
    TestItem,
    TestItemAnnotations,
)
from modelgauge.sut_capabilities import SUTCapability
from modelgauge.tracked_object import TrackedObject
from modelgauge.typed_data import Typeable, TypedData
from typing import Dict, List, Mapping, Sequence, Type


class BaseTest(TrackedObject):
    """This is the placeholder base class for all tests."""

    # Set automatically by @modelgauge_test()
    requires_sut_capabilities: Sequence[Type[SUTCapability]]

    def __init__(self, uid: str):
        super().__init__(uid)
        # The initialization record is set automatically by @modelgauge_test()
        self.initialization_record: InitializationRecord


class PromptResponseTest(BaseTest, ABC):
    """This is the base class for all tests that are single turn."""

    @abstractmethod
    def get_dependencies(self) -> Mapping[str, ExternalData]:
        """Return a mapping of external dependency name to how it can be found downloaded."""
        pass

    @abstractmethod
    def make_test_items(self, dependency_helper: DependencyHelper) -> List[TestItem]:
        """Generate all data that will eventually go to the SUT."""
        pass

    @abstractmethod
    def get_annotators(self) -> Mapping[str, Annotator]:
        """Return a mapping of annotators this Test wants to run.

        Mapping can be empty. Key can be any arbitrary string, and is used to denote
        annotator responses in `measure_quality`.
        """
        pass

    @abstractmethod
    def measure_quality(self, item: TestItemAnnotations) -> Dict[str, float]:
        """Use the SUT responses with annotations to determine how well the SUT did on this TestItem."""
        pass

    @abstractmethod
    def aggregate_measurements(self, items: List[MeasuredTestItem]) -> Typeable:
        """Combine the measurements for each TestItem into a test specific Typeable."""
        pass


class TestResult(TypedData):
    """Container for plugin defined Test result data.

    Every Test can return data however it wants, so this generically
    records the Test's return type and data.
    You can use `to_instance` to get back to the original form.
    """

    # Convince pytest to ignore this class.
    __test__ = False



from modelgauge.annotation import Annotation
from modelgauge.prompt import ChatPrompt, TextPrompt
from modelgauge.sut import SUTCompletion
from modelgauge.typed_data import TypedData
from pydantic import BaseModel, Field
from typing import Dict, List, Mapping, Optional, Type, TypeVar

# TODO: This whole file assumes single turn. We'll either need to make it
# more complicated, or make parallel structures for multi-turn.

_BaseModelType = TypeVar("_BaseModelType", bound=BaseModel)
_Context = TypedData | str | Mapping | None


class PromptWithContext(BaseModel):
    """Combine a prompt with arbitrary context data."""

    prompt: TextPrompt | ChatPrompt
    """The data that goes to the SUT."""

    source_id: Optional[str]
    """Identifier for where this Prompt came from in the underlying datasource."""

    @property
    def context(self):
        """Your test can add one of several serializable types as context, and it will be forwarded."""
        if isinstance(self.context_internal, TypedData):
            return self.context_internal.to_instance()
        return self.context_internal

    context_internal: _Context = None
    """Internal variable for the serialization friendly version of context"""

    def __init__(self, *, prompt, source_id, context=None, context_internal=None):
        if context_internal is not None:
            internal = context_internal
        elif isinstance(context, BaseModel):
            internal = TypedData.from_instance(context)
        else:
            internal = context
        super().__init__(prompt=prompt, source_id=source_id, context_internal=internal)


class TestItem(BaseModel):
    """This is the smallest unit in a Test that can be judged for quality.

    For many Tests, this will be a single Prompt.
    """

    prompts: List[PromptWithContext]

    @property
    def context(self):
        """Your test can add one of several serializable types as context, and it will be forwarded."""
        if isinstance(self.context_internal, TypedData):
            return self.context_internal.to_instance()
        return self.context_internal

    context_internal: _Context = None
    """Internal variable for the serialization friendly version of context"""

    def __init__(self, *, prompts, context=None, context_internal=None):
        if context_internal is not None:
            internal = context_internal
        elif isinstance(context, BaseModel):
            internal = TypedData.from_instance(context)
        else:
            internal = context
        super().__init__(prompts=prompts, context_internal=internal)

    # Convince pytest to ignore this class.
    __test__ = False


class SUTCompletionAnnotations(BaseModel):
    """Pair a SUT's completion with its annotations."""

    completion: SUTCompletion
    annotations: Dict[str, Annotation] = Field(default_factory=dict)
    """All of the annotations, keyed by annotator."""

    def get_annotation(self, key: str, cls: Type[_BaseModelType]) -> _BaseModelType:
        """Convenience function for getting strongly typed annotations."""
        annotation = self.annotations[key]
        return annotation.to_instance(cls)


class SUTResponseAnnotations(BaseModel):
    """All annotated completions for a SUTResponse."""

    completions: List[SUTCompletionAnnotations]


class PromptInteractionAnnotations(BaseModel):
    """Combine a Prompt with the SUT Response to make it easier for Tests to measure quality."""

    prompt: PromptWithContext
    response: SUTResponseAnnotations


class TestItemAnnotations(BaseModel):
    """All of the Interactions with a SUT plus their annotations for a single TestItem."""

    # TODO: This duplicates the list of prompts in the object.
    # Maybe denormalize here.
    test_item: TestItem

    interactions: List[PromptInteractionAnnotations]

    __test__ = False


class MeasuredTestItem(BaseModel):
    """A TestItem with its measurement of quality.

    Note, this does NOT include any SUT Responses or Annotations, as that should already be baked into the Measurements.
    """

    test_item: TestItem
    measurements: Dict[str, float]



from modelgauge.general import get_class
from modelgauge.secret_values import (
    Injector,
    MissingSecretValues,
    RawSecrets,
    Secret,
    SerializedSecret,
)
from typing import Any, Dict, Mapping, Sequence, Tuple


def inject_dependencies(
    args: Sequence[Any], kwargs: Mapping[str, Any], secrets: RawSecrets
) -> Tuple[Sequence[Any], Mapping[str, Any]]:
    """Replace any arg or kwarg injectors with their concrete values."""
    replaced_args = []
    missing_secrets = []
    for arg in args:
        try:
            replaced_args.append(_replace_with_injected(arg, secrets))
        except MissingSecretValues as e:
            missing_secrets.append(e)
        # TODO Catch other kinds of missing dependencies

    replaced_kwargs: Dict[str, Any] = {}
    for key, arg in kwargs.items():
        try:
            replaced_kwargs[key] = _replace_with_injected(arg, secrets)
        except MissingSecretValues as e:
            missing_secrets.append(e)
        # TODO Catch other kinds of missing dependencies
    if missing_secrets:
        raise MissingSecretValues.combine(missing_secrets)

    return replaced_args, replaced_kwargs


def list_dependency_usage(
    args: Sequence[Any], kwargs: Mapping[str, Any], secrets: RawSecrets
) -> Tuple[Sequence[Any], Sequence[Any]]:
    """List all secrets used in the given args and kwargs."""

    def process_item(item):
        """Process an individual item (arg or kwarg)."""
        try:
            replaced_item = _replace_with_injected(item, secrets)
            if isinstance(item, (Injector, SerializedSecret)):
                used_dependencies.append(replaced_item)
        except MissingSecretValues as e:
            missing_dependencies.extend(
                [
                    {
                        "scope": desc.scope,
                        "key": desc.key,
                        "instructions": desc.instructions,
                    }
                    for desc in e.descriptions
                ]
            )
        # TODO Catch other kinds of missing dependencies

    used_dependencies: Sequence[Any] = []
    missing_dependencies: Sequence[Any] = []
    # optional_dependencies: Sequence[Any] = []

    for item in list(args) + list(kwargs.values()):
        process_item(item)

    return used_dependencies, missing_dependencies


def _replace_with_injected(value, secrets: RawSecrets):
    if isinstance(value, Injector):
        return value.inject(secrets)
    if isinstance(value, SerializedSecret):
        cls = get_class(value.module, value.class_name)
        assert issubclass(cls, Secret)
        return cls.make(secrets)
    return value


def serialize_injected_dependencies(
    args: Sequence[Any], kwargs: Mapping[str, Any]
) -> Tuple[Sequence[Any], Mapping[str, Any]]:
    """Replace any injected values with their safe-to-serialize form."""
    replaced_args = []
    for arg in args:
        replaced_args.append(_serialize(arg))
    replaced_kwargs: Dict[str, Any] = {}
    for key, arg in kwargs.items():
        replaced_kwargs[key] = _serialize(arg)
    return replaced_args, replaced_kwargs


def _serialize(arg):
    # TODO Try to make this more generic.
    if isinstance(arg, Secret):
        return SerializedSecret.serialize(arg)
    return arg



from abc import ABC, abstractmethod
from modelgauge.single_turn_prompt_response import PromptWithContext
from modelgauge.sut import SUTCompletion
from pydantic import BaseModel
from typing import Generic, TypeVar

AnnotationType = TypeVar("AnnotationType", bound=BaseModel)


class Annotator(ABC):
    """The base class for all annotators."""

    pass


class CompletionAnnotator(Annotator, Generic[AnnotationType]):
    """Annotator that examines a single prompt+completion pair at a time.

    Subclasses can report whatever class they want, as long as it inherits from Pydantic's BaseModel.
    """

    @abstractmethod
    def translate_request(self, prompt: PromptWithContext, completion: SUTCompletion):
        """Convert the prompt+completion into the native representation for this annotator."""
        pass

    @abstractmethod
    def annotate(self, annotation_request):
        """Perform annotation and return the raw response from the annotator."""
        pass

    @abstractmethod
    def translate_response(self, request, response) -> AnnotationType:
        """Convert the raw response into the form read by Tests."""
        pass



import glob
import os
import shutil
import tempfile
from abc import ABC, abstractmethod
from modelgauge.data_packing import DataUnpacker
from modelgauge.external_data import ExternalData
from modelgauge.general import current_timestamp_millis, hash_file, normalize_filename
from pydantic import BaseModel, Field
from typing import Dict, Mapping, Optional


class DependencyHelper(ABC):
    """Interface for managing the versions of external data dependencies."""

    @abstractmethod
    def get_local_path(self, dependency_key: str) -> str:
        """Return a path to the dependency, downloading as needed."""

    @abstractmethod
    def versions_used(self) -> Mapping[str, str]:
        """Report the version of all dependencies accessed during this run."""

    @abstractmethod
    def update_all_dependencies(self) -> Mapping[str, str]:
        """Ensure the local system has the latest version of all dependencies."""


class DependencyVersionMetadata(BaseModel):
    """Data object we can store along side a dependency version."""

    version: str
    creation_time_millis: int = Field(default_factory=current_timestamp_millis)


class FromSourceDependencyHelper(DependencyHelper):
    """When a dependency isn't available locally, download from the primary source.

    When used, the local directory structure will look like this:
    data_dir/
      dependency_1/
        version_x.metadata
        version_x/
          <dependency's data>
        version_y.metadata
        version_y/
          <dependency's data>
        ...
      dependency_2/
        ...
      ...
    """

    def __init__(
        self,
        data_dir,
        dependencies: Mapping[str, ExternalData],
        required_versions: Mapping[str, str],
    ):
        self.data_dir = data_dir
        self.dependencies = dependencies
        self.required_versions = required_versions
        self.used_dependencies: Dict[str, str] = {}

    def get_local_path(self, dependency_key: str) -> str:
        """Returns the file path, unless the dependency uses unpacking, in which case this returns the directory path."""
        assert dependency_key in self.dependencies, (
            f"Key {dependency_key} is not one of the known "
            f"dependencies: {list(self.dependencies.keys())}."
        )
        external_data: ExternalData = self.dependencies[dependency_key]

        normalized_key = normalize_filename(dependency_key)
        version: str
        if dependency_key in self.required_versions:
            version = self.required_versions[dependency_key]
            self._ensure_required_version_exists(
                dependency_key, normalized_key, external_data, version
            )
        else:
            version = self._get_latest_version(normalized_key, external_data)
        self.used_dependencies[dependency_key] = version
        return self._get_version_path(normalized_key, version)

    def versions_used(self) -> Mapping[str, str]:
        return self.used_dependencies

    def update_all_dependencies(self):
        latest_versions = {}
        for dependency_key, external_data in self.dependencies.items():
            normalized_key = normalize_filename(dependency_key)
            latest_versions[dependency_key] = self._store_dependency(
                normalized_key, external_data
            )
        return latest_versions

    def _ensure_required_version_exists(
        self,
        dependency_key: str,
        normalized_key: str,
        external_data: ExternalData,
        version: str,
    ) -> None:
        version_path = self._get_version_path(normalized_key, version)
        if os.path.exists(version_path):
            return
        # See if downloading from the source creates that version.
        stored_version = self._store_dependency(normalized_key, external_data)
        if stored_version != version:
            raise RuntimeError(
                f"Could not retrieve version {version} for dependency {dependency_key}. Source currently returns version {stored_version}."
            )

    def _get_latest_version(self, normalized_key, external_data) -> str:
        """Use the latest cached version. If none cached, download from source."""
        version = self._find_latest_cached_version(normalized_key)
        if version is not None:
            return version
        return self._store_dependency(normalized_key, external_data)

    def _get_version_path(self, normalized_key: str, version: str) -> str:
        return os.path.join(self.data_dir, normalized_key, version)

    def _find_latest_cached_version(self, normalized_key: str) -> Optional[str]:
        metadata_files = glob.glob(
            os.path.join(self.data_dir, normalized_key, "*.metadata")
        )
        version_creation: Dict[str, int] = {}
        for filename in metadata_files:
            with open(filename, "r") as f:
                metadata = DependencyVersionMetadata.model_validate_json(f.read())
            version_creation[metadata.version] = metadata.creation_time_millis
        if not version_creation:
            return None
        # Returns the key with the max value
        return max(
            version_creation.keys(), key=lambda dict_key: version_creation[dict_key]
        )

    def _store_dependency(self, normalized_key, external_data: ExternalData) -> str:
        with tempfile.TemporaryDirectory() as tmpdirname:
            tmp_location = os.path.join(tmpdirname, normalized_key)
            external_data.download(tmp_location)
            version = hash_file(tmp_location)
            final_path = self._get_version_path(normalized_key, version)
            if os.path.exists(final_path):
                # TODO Allow for overwriting
                return version
            if external_data.decompressor:
                decompressed = os.path.join(
                    tmpdirname, f"{normalized_key}_decompressed"
                )
                external_data.decompressor.decompress(tmp_location, decompressed)
                tmp_location = decompressed
            os.makedirs(os.path.join(self.data_dir, normalized_key), exist_ok=True)
            if not external_data.unpacker:
                shutil.move(tmp_location, final_path)
            else:
                self._unpack_dependency(
                    external_data.unpacker, tmp_location, final_path
                )
            metadata_file = final_path + ".metadata"
            with open(metadata_file, "w") as f:
                f.write(DependencyVersionMetadata(version=version).model_dump_json())
            return version

    def _unpack_dependency(
        self, unpacker: DataUnpacker, tmp_location: str, final_path: str
    ):
        os.makedirs(final_path)
        # TODO Consider if we need to handle the single-file case from HELM.
        unpacker.unpack(tmp_location, final_path)



from contextlib import AbstractContextManager
from threading import Lock
from typing import Generic, TypeVar

T = TypeVar("T")


class ThreadSafeWrapper(AbstractContextManager, Generic[T]):
    """A wrapper that makes thread-hostile objects thread-safe.

    This provides a context manager that holds a lock for accessing the inner object.

    Example usage:

        wrapped_obj = wrapper(thread_hostile_obj)
        with wrapped_obj as obj:
            # Lock is automatically held in here
            obj.do_stuff()
    """

    def __init__(self, wrapped: T):
        self._wrapped = wrapped
        self._lock = Lock()

    def __enter__(self) -> T:
        self._lock.__enter__()
        return self._wrapped

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self._lock.__exit__(exc_type, exc_value, traceback)



"""
This namespace plugin loader will discover and load all plugins from modelgauge's plugin directories.

To see this in action:

* poetry install
* poetry run modelgauge list
* poetry install --extras demo
* poetry run modelgauge list

The demo plugin modules will only print on the second run.
"""

import importlib

from tqdm import tqdm
import modelgauge
import modelgauge.annotators
import modelgauge.runners
import modelgauge.suts
import modelgauge.tests
import pkgutil
from types import ModuleType
from typing import Iterator, List


def _iter_namespace(ns_pkg: ModuleType) -> Iterator[pkgutil.ModuleInfo]:
    return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")


def list_plugins() -> List[str]:
    """Get a list of plugin module names without attempting to import them."""
    module_names = []
    for ns in ["tests", "suts", "runners", "annotators"]:
        for _, name, _ in _iter_namespace(getattr(modelgauge, ns)):
            module_names.append(name)
    return module_names


def load_plugins(disable_progress_bar: bool = False) -> None:
    """Import all plugin modules."""
    plugins = list_plugins()
    for module_name in tqdm(
        plugins,
        desc="Loading plugins",
        disable=disable_progress_bar or len(plugins) == 0,
    ):
        importlib.import_module(module_name)



from modelgauge.general import get_class
from pydantic import BaseModel
from typing import Any, Dict, Optional, Type, TypeVar
from typing_extensions import Self

Typeable = BaseModel | Dict[str, Any]

_BaseModelType = TypeVar("_BaseModelType", bound=Typeable)


def is_typeable(obj) -> bool:
    """Verify that `obj` matches the `Typeable` type.

    Python doesn't allow isinstance(obj, Typeable).
    """
    if isinstance(obj, BaseModel):
        return True
    if isinstance(obj, Dict):
        for key in obj.keys():
            if not isinstance(key, str):
                return False
        return True
    return False


class TypedData(BaseModel):
    """This is a generic container that allows Pydantic to do polymorphic serialization.

    This is useful in situations where you have an unknown set of classes that could be
    used in a particular field.
    """

    module: str
    class_name: str
    data: Dict[str, Any]

    @classmethod
    def from_instance(cls, obj: Typeable) -> Self:
        """Convert the object into a TypedData instance."""
        if isinstance(obj, BaseModel):
            data = obj.model_dump()
        elif isinstance(obj, Dict):
            data = obj
        else:
            raise TypeError(f"Unexpected type {type(obj)}.")
        return cls(
            module=obj.__class__.__module__,
            class_name=obj.__class__.__qualname__,
            data=data,
        )

    def to_instance(
        self, instance_cls: Optional[Type[_BaseModelType]] = None
    ) -> _BaseModelType:
        """Convert this data back into its original type.

        You can optionally include the desired resulting type to get
        strong type checking and to avoid having to do reflection.
        """
        cls_obj: Type[_BaseModelType]
        if instance_cls is None:
            cls_obj = get_class(self.module, self.class_name)
        else:
            cls_obj = instance_cls
        assert (
            cls_obj.__module__ == self.module
            and cls_obj.__qualname__ == self.class_name
        ), (
            f"Cannot convert {self.module}.{self.class_name} to "
            f"{cls_obj.__module__}.{cls_obj.__qualname__}."
        )
        if issubclass(cls_obj, BaseModel):
            return cls_obj.model_validate(self.data)  # type: ignore
        elif issubclass(cls_obj, Dict):
            return cls_obj(self.data)  # type: ignore
        else:
            raise TypeError(f"Unexpected type {cls_obj}.")
